/*
 *   xsnap -- take a snapshot of a portion of the screen.
 *
 *   Copyright 1989 Clauss Strauch
 *                  cbs@cad.cs.cmu.edu
 *
 *   Permission to use, copy, modify, and distribute this software and its
 *   documentation for any purpose and without fee is hereby granted.
 *   This software is provided "as is", without express or implied warranty.
 *   
 *   Modified by Bill Janssen janssen@parc.xerox.com 
 *    add -xwd and -region options
 *   Modified by Arnaud Le Hors lehors@mirsa.inria.fr  November 14 1990
 *    fix code about region option
 *    add -xpm option
 *   Modified by Jean-Pierre Demailly demailly@fourier.ujf-grenoble.fr :
 *    added -jpg, -png, -ppm, -pnm formats, compression, and GUI dialog.
 */

#include <stdio.h> 
#include <unistd.h>
#include <sys/stat.h>
#include <zlib.h>
#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include "patchlevel.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

/*  Leave arguments as globals, since there are so many of them.
 *  They'll only be referenced in process_args and main.
 */

enum { JPG_F=0, PNG_F, PNM_F, PPM_F, XPM_F, XWD_F, NUM_F };
enum { BD_COLOR=0, BG_COLOR, FG_COLOR, 
       BT1_COLOR, BT2_COLOR, BT3_COLOR, BT4_COLOR, NUM_COLORS };

static char * suffix[2][NUM_F] = {
  { ".jpg", ".png", ".pnm", ".ppm", ".xpm", ".xwd" },
  { ".jpg", ".png", ".pnm.gz", ".ppm.gz", ".xpm.gz", ".xwd.gz" }
};

char *color_string[NUM_COLORS] = {
  NULL,				     /* name of border color */
  "#DCDAD5",                         /* name of background color */
  "black",                           /* name of foreground color */
  "black",                           /* name of 1st button color */
  "grey50",                          /* name of 2nd button color */
  "grey95",                          /* name of 3rd button color */
  "white"                            /* name of 4th button color */
};
unsigned long color_pixel[NUM_COLORS]; /* pixels */

char *display_string = NULL;         /* display we'll open */
int  border_width = 2;               /* border_width  of the snapshot window */
int  start_iconic = FALSE;           /* start snap window in iconic state?  */
char *window_geom_string = NULL;     /* geometry of snapshot window */
char *icon_geom_string = NULL;       /* icon geometry */
char *region_geom_string = NULL;     /* location of region to copy */
char win_name0[128]="";              /* name of window snapshot */
char win_name[256]="";               /* name of window snapshot */
#ifdef PNG
int  output_fmt = PNG_F;	             /* output format */
#else
int  output_fmt = PPM_F;	             /* output format */
#endif
char output_name[256] = "";          /* file to store in */
char root_name[256] = "?";	     /* root name of file */
int  grab_server = TRUE;             /* grab the server? */
int  display_in_window = TRUE;       /* display in window after snap? */
int  save_in_file = FALSE;           /* save in file after snap? */
int  zip = TRUE;                     /* use compression */
int  clicked =-1;                    /* button pressed */

Colormap private_cmap = 0;

/*
 *  create_event_window returns the ID of a InputOnly window that covers
 *  the given window.
 */

Window create_event_window(dpy,win, init_cursor)
Display *dpy;
Window win;
{
	XSetWindowAttributes xswa;
	unsigned long xswa_valuemask;
	Window root_win;
	Window event_window;
	unsigned int win_width, win_height;
	unsigned int  win_border_width, win_depth;
	int win_x, win_y;

	/* get the geometry of the window  */

	XGetGeometry(dpy, win, &root_win, &win_x, &win_y,
	    &win_width, &win_height, &win_border_width, &win_depth);

	/* make an input only window to get events from  */

	xswa.cursor = init_cursor;
	xswa.override_redirect = True;
	xswa.event_mask = ButtonPressMask | ButtonReleaseMask | Button1MotionMask;
	xswa_valuemask = CWCursor | CWOverrideRedirect | CWEventMask;
	event_window = XCreateWindow(dpy, win, win_x, win_y, win_width,
	    win_height, 0, 0, InputOnly, CopyFromParent, 
	    xswa_valuemask, &xswa);
	return(event_window);
}

/*
 *   draw box draws a box on the given window, with the given GC 
 *
 */

void draw_box(dpy, win, a_gc, x1, y1, x2, y2)
GC a_gc;
Window win;
Display *dpy;
int x1, y1, x2, y2;
{
	XSegment segments[4];
	segments[0].x1 = (short)x1;
	segments[0].y1 = (short)y1;
	segments[0].x2 = (short)x1;
	segments[0].y2 = (short)y2;

	segments[1].x1 = (short)x1;
	segments[1].y1 = (short)y1;
	segments[1].x2 = (short)x2;
	segments[1].y2 = (short)y1;

	segments[2].x1 = (short)x2;
	segments[2].y1 = (short)y2;
	segments[2].x2 = (short)x1;
	segments[2].y2 = (short)y2;

	segments[3].x1 = (short)x2;
	segments[3].y1 = (short)y2;
	segments[3].x2 = (short)x2;
	segments[3].y2 = (short)y1;

	XDrawSegments(dpy, win, a_gc, segments, 4);
}

/*
 *  get_region
 * takes as input:
 *    display
 *    window to get region from 
 *    pointers to x1, y1, width, height
 *  
 *   returns:  the position and width and height of a
 *             user selected region via the given pointers.
 *
 */

void
find_window(dpy, bool, x, y, u, v, width, height)
     Display *dpy;
     int bool;
     int x, y;	  
     int *u, *v, *width, *height;
{
  XWindowAttributes wa;
  Window findW = DefaultRootWindow(dpy), stopW=0, childW, initW;

  XTranslateCoordinates(dpy, findW, findW, x, y, &x, &y, &stopW);

  if (stopW) 
    initW = stopW;
  else
    initW = findW;

  while (stopW) {
    XTranslateCoordinates(dpy, findW, stopW, x, y, &x, &y, &childW);
    findW = stopW;
    if (childW &&
	XGetWindowAttributes(dpy, childW, &wa) &&
	wa.class != InputOutput)
	break;
    stopW = childW;
    }

  XGetWindowAttributes(dpy, findW, &wa);
  *width = wa.width;
  *height = wa.height;
  private_cmap = wa.colormap;

  if (!bool) findW = initW;

  XTranslateCoordinates(dpy, findW, DefaultRootWindow(dpy), 0, 0, u, v,&stopW);
}

get_region(dpy, win,  x, y, width, height)
Display *dpy;
Window win;
int *x, *y;
unsigned int *height, *width;

{
	Window event_window;
	Cursor up_right_curs, up_left_curs;
	Cursor low_right_curs, low_left_curs;
	Cursor current_cursor;
	int done;
	int init_x, init_y;
	int last_x, last_y;
	XEvent event;
	GC xor_gc;
	XGCValues xor_gc_values;             /* for creating xor_gc */
	unsigned long xor_gc_valuemask;       /* valuemask for creating xor_gc */

	/* make the GC and cursors we'll need */

	up_right_curs = XCreateFontCursor(dpy, XC_ur_angle);

	up_left_curs = XCreateFontCursor(dpy, XC_ul_angle);

	low_right_curs = XCreateFontCursor(dpy, XC_lr_angle);
	
	low_left_curs = XCreateFontCursor(dpy, XC_ll_angle);
	


	xor_gc_valuemask = GCFunction | GCSubwindowMode  | GCForeground;
	xor_gc_values.function = GXxor;
	xor_gc_values.foreground = 0xfd;
	xor_gc_values.subwindow_mode = IncludeInferiors;
	xor_gc = XCreateGC(dpy, win, xor_gc_valuemask, &xor_gc_values);

	event_window = create_event_window(dpy, win,up_left_curs);
	XMapRaised(dpy, event_window);

	if (XGrabPointer(dpy, event_window, True, 
	    ButtonPressMask,
	    GrabModeAsync, GrabModeAsync, None, up_left_curs, 
			 CurrentTime) != 0)
	  {
	    fprintf(stderr, "Cannot grab pointer.");
	    exit(1);
	  }


	/* get the initial button  press */
	done = 0;
	while (done == 0)
	{
	        XNextEvent(dpy, &event);
		switch(event.type)
		{
		case MappingNotify:
			XRefreshKeyboardMapping((XMappingEvent *)&event);
			break;
		case ButtonPress:
			if (event.xbutton.button == 1)
			{
				init_x = event.xbutton.x;
				init_y = event.xbutton.y;
				done = 1;
				break;
			}
			if (event.xbutton.button == 2)
			{
				init_x = event.xbutton.x;
				init_y = event.xbutton.y;
				done = 2;
				break;
			}
			if (event.xbutton.button == 3) 
			  {
			    if (grab_server)
			      {
				XGrabServer(dpy);
				XSync(dpy, 0);
			      }
			    exit(0);
			  }

		}
	}


	/*  now we have the location of one corner of the box.   change the cursor,
 *  and have the user drag out the area.
 */
	last_x = init_x;
	last_y = init_y;
        if (done == 1)
	  {
	  current_cursor = low_right_curs;
	  XChangeActivePointerGrab(dpy, ButtonReleaseMask | Button1MotionMask,
	    current_cursor, CurrentTime);
	  done = 0;
 	  draw_box(dpy, win, xor_gc, init_x, init_y, last_x, last_y);
	  }
	while (! done)
	{
		XNextEvent(dpy, &event);
		switch(event.type)
		{
		case MappingNotify:
			XRefreshKeyboardMapping((XMappingEvent *)&event);
			break;
		case MotionNotify:
			draw_box(dpy, win, xor_gc, 
			    init_x, init_y, last_x, last_y);  /* erase old */
			last_x = event.xmotion.x;
			last_y = event.xmotion.y;
			draw_box(dpy, win, xor_gc, 
			    init_x, init_y, last_x, last_y); /* draw new  */
			/*  Change cursor to correspond to position of pointer */
			if ((init_x < last_x) && (init_y < last_y)
			    && (current_cursor != low_right_curs))
			{
				current_cursor = low_right_curs;
				XChangeActivePointerGrab(dpy, 
				    ButtonReleaseMask | Button1MotionMask,
				    low_right_curs, CurrentTime);
			}
			else if ((last_x < init_x) && (last_y < init_y)
			    &&  (current_cursor != up_left_curs))
			{
				current_cursor = up_left_curs;
				XChangeActivePointerGrab(dpy, 
				    ButtonReleaseMask | Button1MotionMask,
				    up_left_curs, CurrentTime);
			}

			else if ((init_x < last_x) && (last_y < init_y)
			    && (current_cursor != up_right_curs))
			{
				current_cursor = up_right_curs;
				XChangeActivePointerGrab(dpy, 
				    ButtonReleaseMask | Button1MotionMask,
				    up_right_curs, CurrentTime);

			}
			else if ((last_x < init_x) && (init_y < last_y)
			    && (current_cursor != low_left_curs))
			{
				current_cursor = low_left_curs;
				XChangeActivePointerGrab(dpy, 
				    ButtonReleaseMask | Button1MotionMask,
				    low_left_curs, CurrentTime);
			}

			break;
		case ButtonRelease:
			if (event.xbutton.button == 1)
			{
				done = TRUE;
				draw_box(dpy, win, xor_gc, 
				    init_x, init_y, last_x, last_y);  /* erase last box drawn */
			}
			break;
		}
	}
	XFlush(dpy);   /*  gets rid of last box on screen  */
	if (init_x < last_x) *x = init_x;
	else *x = last_x;
	if (init_y < last_y) *y = init_y;
	else *y = last_y;
	*width = (unsigned int)abs(last_x - init_x);
	*height = (unsigned int)abs(last_y - init_y);
	/* clean up after ourself: */

	XDestroyWindow(dpy, event_window);
	XFreeGC(dpy, xor_gc);

	/* we'll let the caller ungrab the pointer */

        if (done == 1)
	  {
	  int u;
          if (*width==0 && *height==0)
             find_window(dpy, 0, init_x,init_y, x, y, width, height);
	  else
 	     find_window(dpy, 0, init_x, init_y, &u, &u, &u, &u);
	  }

        if (done == 2)
          find_window(dpy, 1, init_x,init_y, x, y, width, height);
}

/* 
 *  get_pixmap_region 
 *
 *       input :
 *               a display, a window, x, y, width, height, interactive.
 *               if interactive, the user is prompted for a region,
 *               other wise the given region is copied to a pixmap.
 *       returns : a pixmap containing a copy of a user-specified area
 *                 of the given window;
 *
 */

Pixmap get_pixmap_region(dpy, a_screen, win, a_gc,  x, y, 
width, height, depth, interactive)
Display *dpy;
GC a_gc;
int a_screen;
Window win;
int *x, *y;
unsigned int *width, *height, *depth;
int interactive;

{
	int reg_x, reg_y;
	unsigned int reg_width, reg_height;
	Pixmap pixmap_returned;
	int return_flag;
        int junk_left, junk_top, junk_width, junk_height, junk_border_width;
	Window junk_root;

	if (interactive){
		get_region(dpy, win, &reg_x, &reg_y,
		    &reg_width, &reg_height);
		*x = reg_x;
		*y = reg_y;
		*width = reg_width;
		*height = reg_height;
	}

        if (*width==0 || *height==0) exit(1);

	/* Use the depth of `win' for the depth of the pixmap */
	  
	XGetGeometry (dpy, win, &junk_root, &junk_left, &junk_top,
		      &junk_width, &junk_height, &junk_border_width, depth);

	pixmap_returned = XCreatePixmap(dpy, 
					DefaultRootWindow(dpy),
					*width, *height, *depth);

	/*  now copy the area we specified  */

	XCopyArea(dpy, win, pixmap_returned, a_gc, *x, *y, 
		  *width, *height, 0, 0);
	if (interactive)  XUngrabPointer(dpy, CurrentTime);
	return(pixmap_returned);
}


/* 
 *  X error handler for,  mainly for future use
 */

int my_X_error_handler(display, myerr)
Display *display;
XErrorEvent *myerr;
{
  char msg[80];  

/* if we get a BadAlloc error, it's probably 'cuz we couldn't allocate
 * the pixmap.  Tell the user this.
 */
  fprintf(stderr,"xsnap:\n"); 
  if ((myerr->error_code) == BadAlloc)   
    {
      fprintf(stderr, "Insufficient resources for operation.\n");
      fprintf(stderr, "Probably not enough memory to allocate pixmap.\n");
    }

/* Now print out the error string that X generates */
  
  XGetErrorText(display, myerr->error_code, msg,80);
  fprintf(stderr, "Error code %s\n", msg);
/* no way to continue, so we'll just bag it  */
  exit(1);
}
  

/*
 *  print a usage message to the user
 */

char * usage_msg[] = {
"Usage:  xsnap [-options...]",
"",
"where options are:",
"\t[-display host:display] Display to connect to.",
"\t[-bg bgcolor]           Background color",
"\t[-fg bgcolor]           Foreground color",
"\t[-bd bordercolor]       Color of border",
"\t[-bw borderwidth]       Width of border",
"\t[-geometry geom]        Geometry of snapshot window",
"\t[-icongeometry geom]    Location of icon.",
"\t[-region geom]          Region of screen to be copied.",
"\t[-iconic]               Start up in iconic state",
"\t[-name name]            Passed to window manager for name of window.",
"\t[-nograb]               Don't grab server during selection of region.",
"\t[-jpg]                  Write output in 'jpg' format.",
"\t[-png]                  Write output in 'png' format.",
"\t[-pnm]                  Write output in 'pnm(.gz)' format.",
"\t[-ppm]                  Write output in 'ppm(.gz)' format.",
"\t[-xpm]                  Write output in 'xpm(.gz)' format.",
"\t[-xwd]                  Write output in 'xwd(.gz)' format.",
"\t[-dump]                 Dump to file ($HOME/snapXYZT.<fmt> by default).",
"\t[-file filename]        Write output to file 'filename'.",
"\t[-stdout]               Write output to stdout.",
"\t[-show]                 Display in window (default).",
"\t[-hide]                 Don't display in window.",
"\t[-nozip]                Don't use compression (compression is default).",
""
};

	
char *bindings_msg[] = {
"Button bindings at runtime:",
"\tMouse button 1: select region of capture, or framed window",
"\tMouse button 2: click on window to be captured",
"\tMouse button 3: cancel snapshot",
"",
"Key bindings at runtime:",
"\tq/Q\texit xsnap",
"\tz/Z\tenable/disable compression",
"\tj/J\tsave snapshot as .jpg file",
"\tg/G\tsave snapshot as .png file",
"\tn/N\tsave snapshot as .pnm(.gz) file",
"\tp/P\tsave snapshot as .ppm(.gz) file",
"\tx/X\tsave snapshot as .xpm(.gz) file",
"\tw/W\tsave snapshot as .xwd(.gz) file",
"\t ! \trestart new xsnap instance",
"\tother\tshow menu and key bindings."
};
   
usage()
{
  int i;
	printf("\nXsnap, version %s.%s\n\n", VERSION, PATCHLEVEL);
        for (i=0; i<sizeof(usage_msg)/sizeof(char *); i++)
	    printf("%s\n", usage_msg[i]);
        for (i=0; i<sizeof(bindings_msg)/sizeof(char *); i++)
            printf("%s\n",bindings_msg[i]);
	exit(1);
}

/*
 *        get the arguments.   arguments are matched to the first
 *        distinguishing substring.
 */
process_args(argc, argv)
int argc;
char **argv;
{
	int i;

	for (i = 1; i < argc; i++)
	{
		if (strncmp(argv[i], "-h", 2) == 0)
		{
			usage();
			continue;
		}
		if (strncmp(argv[i], "-disp", 5) == 0)
		{
			display_string = argv[++i];
			continue;
		}
		if (strncmp(argv[i], "-bw", 3) == 0)
		{
			border_width = atoi(argv[++i]);
			continue;
		}
		if (strncmp(argv[i], "-bg", 3) == 0)
		{
			color_string[BG_COLOR] = argv[++i];
			continue;
		}
		if (strncmp(argv[i], "-fg", 3) == 0)
		{
			color_string[FG_COLOR] = argv[++i];
			continue;
		}
		if (strncmp(argv[i], "-bd", 3) == 0)
		{
			color_string[BD_COLOR] = argv[++i];
			continue;
		}
		if (strncmp(argv[i], "-bt", 3) == 0 &&
                    argv[i][3]>='1' && argv[i][3]<='4')
		{
			color_string[BT1_COLOR+argv[i][3]-'1'] = argv[i+1];
			++i;
			continue;
		}
		if (strncmp(argv[i], "-geom", 5) == 0)
		{
			window_geom_string = argv[++i];
			continue;
		}
		if (strncmp(argv[i], "-icongeom", 9) == 0)
		{
			icon_geom_string = argv[++i];
			continue;
		}
		if (strncmp(argv[i], "-region", 7) == 0)
		{
			region_geom_string = argv[++i];
			continue;
		}
		if (strncmp(argv[i], "-iconic", 7) == 0)
		{
			start_iconic = TRUE;
			continue;
		}
		if (strncmp(argv[i], "-name", 5) == 0)
		{
			strncpy(win_name0 , argv[++i], 126);
			win_name0[127] = '\0';
			continue;
		}
		if (strncmp(argv[i], "-nograb", 7) == 0)
		{
			grab_server = FALSE;
			continue;
		}
		if (strncmp(argv[i], "-hide", 5) == 0)
		{
		  display_in_window = FALSE;
		  continue;
		}
		if (strncmp(argv[i], "-show", 5) == 0)
		{
		  display_in_window = TRUE;
		  continue;
		}
		if (strncmp(argv[i], "-dump", 5) == 0)
		{
		  save_in_file = TRUE;
		  continue;
		}
		if (strncmp(argv[i], "-nozip", 5) == 0)
		{
		  zip = FALSE;
		  continue;
		}
		if (strncmp(argv[i], "-jpg", 4) == 0)
		{
		  output_fmt = JPG_F;
		  continue;
		}
		if (strncmp(argv[i], "-png", 4) == 0)
		{
		  output_fmt = PNG_F;
		  continue;
		}
		if (strncmp(argv[i], "-pnm", 4) == 0)
		{
		  output_fmt = PNM_F;
		  continue;
		}
		if (strncmp(argv[i], "-ppm", 4) == 0)
		{
		  output_fmt = PPM_F;
		  continue;
		}
		if (strncmp(argv[i], "-xpm", 4) == 0)
		{
		  output_fmt = XPM_F;
		  continue;
		}
		if (strncmp(argv[i], "-xwd", 4) == 0)
		{
		  output_fmt = XWD_F;
		  continue;
		}
		if (strncmp(argv[i], "-stdout", 7) == 0)
		{
                  *root_name='\0';
                  display_in_window=FALSE;
                  save_in_file = TRUE;
		  continue;
		}
		if (strncmp(argv[i], "-file", 5) == 0)
		{
		  if (i<argc-1) 
                     strcpy(root_name, argv[++i]);
                  display_in_window=FALSE;
                  save_in_file = TRUE;
		  continue;
		}
		usage();
	}
	if (!display_in_window) save_in_file = TRUE;
}

static char err_malloc[] = "xsnap: out of memory trying to allocate %s\n";

/*
 * Get the XColors of all pixels in image - returns # of colors
 */
int Get_XColors(dpy, win_info, colors)
     Display *dpy;
     XWindowAttributes *win_info;
     XColor **colors;
{
    int i, ncolors;

    if (!win_info->colormap)
	return(0);

    if (win_info->visual->class == TrueColor ||
	win_info->visual->class == DirectColor)
	return(0);    /* XXX punt for now */

    ncolors = win_info->visual->map_entries;
    if (!(*colors = (XColor *) malloc (sizeof(XColor) * ncolors)))
      fprintf (stderr, err_malloc, "colors");

    for (i=0; i<ncolors; i++)
      (*colors)[i].pixel = i;

    XQueryColors(dpy, win_info->colormap, *colors, ncolors);
    
    return(ncolors);
}

/*
 * Determine the pixmap size.
 */

int Image_Size(image, format)
     XImage *image;
     int format;
{
    if (format != ZPixmap)
      return(image->bytes_per_line * image->height * image->depth);

    return(image->bytes_per_line * image->height);
}

_swapshort (bp, n)
    register char *bp;
    register unsigned n;
{
    register char c;
    register char *ep = bp + n;

    while (bp < ep) {
	c = *bp;
	*bp = *(bp + 1);
	bp++;
	*bp++ = c;
    }
}

_swaplong (bp, n)
    register char *bp;
    register unsigned n;
{
    register char c;
    register char *ep = bp + n;
    register char *sp;

    while (bp < ep) {
	sp = bp + 3;
	c = *sp;
	*sp = *bp;
	*bp++ = c;
	sp = bp + 1;
	c = *sp;
	*sp = *bp;
	*bp++ = c;
	bp += 2;
    }
}

check_filename()
{
int i, j, found;
struct stat buf;
char *suf;
      suf = suffix[zip][output_fmt];
      i=0;
      found=1;
      if (*root_name=='?')
      while(found && i<10000) {
	found = 0;
	j = zip;
	do {
	    j = 1-j;
	    sprintf(output_name,"%s/snap%04d%s",
               ((getenv("HOME"))?(char *)getenv("HOME") : P_tmpdir), i, 
		    suffix[j][output_fmt]);
            found |= !(-stat(output_name, &buf));	   
	} while (j!=zip);
        ++i;
      } else {
        strcpy(output_name, root_name);
	if (!strstr(output_name, suf) ||
            strcmp(output_name+(strlen(output_name)-strlen(suf)),suf)) 
 	  strcat(output_name, suf);
      }
}

#include "X11/XWDFile.h"

save_as_xwd_file (display, screen, win_info, pixmap, width, height, depth)
     Display *display;
     int screen;
     XWindowAttributes *win_info;
     Pixmap pixmap;
     unsigned int width, height, depth;
{
    unsigned long swaptest = 1;
    XColor *colors;
    unsigned buffer_size;
    int win_name_size;
    int header_size;
    int ncolors, i;
    char *win_name;
    XImage *image;
    int absx, absy, x, y;
    int dwidth, dheight;
    int bw;
    XWDFileHeader header;
    gzFile *output;

    image = XGetImage (display, pixmap, 0, 0, width, height, AllPlanes, ZPixmap);

    if (!image) {
	fprintf (stderr, "xsnap:  unable to get image at %dx%d+%d+%d\n",
		 width, height, 0, 0);
	exit (1);
    }

    /*
     * Find the output file
     */

    if (*root_name == '\0')
      output = NULL;
    else
      {
      check_filename();
      output = gzopen (output_name, (zip)?"wb6":"wb0");
      }

    /*
     * Determine the pixmap size.
     */
    buffer_size = Image_Size(image, ZPixmap);

    win_info->colormap = private_cmap;
    ncolors = Get_XColors(display, win_info, &colors);

    /*
     * Calculate header size.
     */
    header_size = sizeof(header);

    /*
     * Write out header information.
     */
    header.header_size = (xwdval) header_size + 4;
    header.file_version = (xwdval) XWD_FILE_VERSION;
    header.pixmap_format = (xwdval) ZPixmap;
    header.pixmap_depth = (xwdval) image->depth;
    header.pixmap_width = (xwdval) image->width;
    header.pixmap_height = (xwdval) image->height;
    header.xoffset = (xwdval) image->xoffset;
    header.byte_order = (xwdval) image->byte_order;
    header.bitmap_unit = (xwdval) image->bitmap_unit;
    header.bitmap_bit_order = (xwdval) image->bitmap_bit_order;
    header.bitmap_pad = (xwdval) image->bitmap_pad;
    header.bits_per_pixel = (xwdval) image->bits_per_pixel;
    header.bytes_per_line = (xwdval) image->bytes_per_line;
    header.visual_class = (xwdval) win_info->visual->class;
    header.red_mask = (xwdval) win_info->visual->red_mask;
    header.green_mask = (xwdval) win_info->visual->green_mask;
    header.blue_mask = (xwdval) win_info->visual->blue_mask;
    header.bits_per_rgb = (xwdval) win_info->visual->bits_per_rgb;
    header.colormap_entries = (xwdval) win_info->visual->map_entries;
    header.ncolors = ncolors;
    header.window_width = (xwdval) win_info->width;
    header.window_height = (xwdval) win_info->height;
    header.window_x = 0;
    header.window_y = 0;
    header.window_bdrwidth = (xwdval) win_info->border_width;

    if (*(char *) &swaptest) {
      _swaplong((char *) &header, sizeof(header));
      for (i = 0; i < ncolors; i++) {
        _swaplong((char *) &colors[i].pixel, sizeof(long));
        _swapshort((char *) &colors[i].red, 3 * sizeof(short));
      }
    }

    if (output) {
       gzwrite(output, (char *)&header, sizeof(header));
       gzwrite(output, "xwd\0", 4);
    } else {
       fwrite((char *)&header, sizeof(header), 1, stdout);
       fwrite("xwd\0", 4, 1, stdout);
    }

    /*
     * Write out the color maps, if any
     */
    if (output)
       gzwrite(output, (char *) colors, sizeof(XColor)*ncolors);
    else
       fwrite((char *) colors, sizeof(XColor), ncolors, stdout);

    /*
     *    This copying of the bit stream (data) to a file is to be replaced
     *  by an Xlib call which hasn't been written yet.  It is not clear
     *  what other functions of xwd will be taken over by this (as yet)
     *  non-existant X function.
     */
    if (output)
       gzwrite(output, image->data, (int) buffer_size);
    else
       fwrite(image->data, (int) buffer_size, 1, stdout);

    /*
     * close the file
     */
    if (output)
      gzclose (output);

    /*
     * free the color buffer.
     */
    if(ncolors > 0) free(colors);

    /*
     * Free image
     */
    XDestroyImage(image);
}

void scan_ximage_line(char *scanp, char *ximp, int width, int depth)
{
int i;
unsigned char u, v;

  if (depth>0) {
    if (depth==8)
	memcpy(scanp, ximp, width);
    else
    if (depth==15)
        for (i=0; i<width; i++) {
	    u = *ximp++;
            v = (*ximp++)&127;
	    *scanp++ = ((v>>2)*255)/31;
	    *scanp++ = ((((v&3)<<3) | (u>>5))*255)/31;
	    *scanp++ = ((u&31)*255)/31;
        }
    else
    if (depth==16)
        for (i=0; i<width; i++) {
	    u = *ximp++;
            v = *ximp++;
	    *scanp++ = ((v>>3)*255)/31;
	    *scanp++ = ((((v&7)<<3) | (u>>5))*255)/63;
	    *scanp++ = ((u&31)*255)/31;
        }
    else
    if (depth==24)
        for (i=0; i<width; i++) {
	    scanp[2] = *ximp++;
	    scanp[1] = *ximp++;
	    scanp[0] = *ximp++;
            scanp+=3;
        }
    else
    if (depth==32)
        for (i=0; i<width; i++) {
	    scanp[2] = *ximp++;
	    scanp[1] = *ximp++;
	    scanp[0] = *ximp++;
	    ximp++;
            scanp+=3;
        } 
  } else {
    if (depth==-8)
	memcpy(scanp, ximp, width);
    else
    if (depth==-15)
        for (i=0; i<width; i++) {
	    v = *ximp++;
            u = (*ximp++)&127;
	    *scanp++ = ((v>>2)*255)/31;
	    *scanp++ = ((((v&3)<<3) | (u>>5))*255)/31;
	    *scanp++ = ((u&31)*255)/31;
        }
    else
    if (depth==-16)
        for (i=0; i<width; i++) {
	    v = *ximp++;
            u = *ximp++;
	    *scanp++ = ((v>>3)*255)/31;
	    *scanp++ = ((((v&7)<<3) | (u>>5))*255)/63;
	    *scanp++ = ((u&31)*255)/31;
        }
    else
    if (depth==-24)
        memcpy(scanp, ximp, 3*width);
    else
    if (depth==-32)
        for (i=0; i<width; i++) {
	    ximp++;
	    *scanp++ = *ximp++;
	    *scanp++ = *ximp++;
	    *scanp++ = *ximp++;
        } 
  }
}
	   
save_as_ppm_file (display, screen, win_info, pixmap, width, height, depth)
     Display *display;
     int screen;
     XWindowAttributes *win_info;
     Pixmap pixmap;
     unsigned int width, height, depth;
{
    int           i, j, color_depth, ncolors, raw;
    int           bit_depth=0, color_type=0, num_palette, rowbytes;
    unsigned char * palette = NULL;
    unsigned char *scanp, *ximp;
    char          outstr[60];
    Visual        *visual;
    XColor *colors;
    XImage *image;
    gzFile *output;

    image = XGetImage (display, pixmap, 0,0, width,height, AllPlanes, ZPixmap);
    color_depth = image->bits_per_pixel;
    visual = DefaultVisual(display, screen);
    if (color_depth == 16 && visual->green_mask==992) color_depth = 15;
    if (ImageByteOrder(display) == MSBFirst) color_depth = -color_depth;

    if (!image) {
        fprintf (stderr, "xsnap:  unable to get image at %dx%d+%d+%d\n",
		 width, height, 0, 0);
	return;
    }

    /*
     * Find the output file
     */

    if (*root_name == '\0')
      output = NULL;
    else
      {
      check_filename();
      output = gzopen (output_name, (zip)?"wb6":"wb0");
      if (!output) return;
      }

    if (image->depth == 8) {
        char *ptr;
        win_info->colormap = private_cmap;
        ncolors = Get_XColors(display, win_info, &colors);
        palette = ptr = (unsigned char *)malloc(3*ncolors);
        if (!palette) {
	  fprintf (stderr, err_malloc, "palette");
	  if (output)
              gzclose(output);
           return;
	}
        for (i=0;i<ncolors;i++) {
	   *ptr++ = colors[i].red>>8;
	   *ptr++ = colors[i].green>>8;
	   *ptr++ = colors[i].blue>>8;
	}
    }
    scanp = (unsigned char *)malloc(3*image->width);
    if (!scanp) {
        fprintf (stderr, err_malloc, "scanline");        
        free(palette);
	if (output)
            gzclose(output);
        return;
    }
    
    raw = (output_fmt == PPM_F);
    sprintf(outstr, "P%d\n# Creator xsnap %s.%s\n%d %d\n255\n",
            3+3*raw, VERSION, PATCHLEVEL, image->width, image->height);
    if (output)
      gzwrite(output, outstr, strlen(outstr));
    else
      fprintf(stdout, "%s", outstr);

    for (j = 0;  j < image->height;  ++j) {
        ximp = image->data + j * image->bytes_per_line;
	if (image->depth==8) {
	   i = image->width-1;
           while(i>=0) {
              memcpy(&scanp[3*i], &palette[3*ximp[i]], 3);
              --i;
	   }
 	} else
           scan_ximage_line(scanp, ximp, image->width, color_depth);
       for (i=0; i<image->width; i++) {
	    if (raw) {
	        sprintf(outstr, "%c%c%c\0", 
			scanp[3*i], scanp[3*i+1], scanp[3*i+2]);
                if (output)
                    gzwrite(output, outstr, 3);
                else
                    fwrite(outstr, 3, 1, stdout);
            } else {
	        sprintf(outstr, "%03d %03d %03d %s", 
		        (int)scanp[3*i], (int)scanp[3*i+1], (int)scanp[3*i+2],
                        ((i==image->width-1)?"\n":
	                 ((i%6==5)?"\n  ":" ")));
                if (output)
                    gzwrite(output, outstr, strlen(outstr));
                else
                    fprintf(stdout, "%s", outstr);
	    }
	}
    }
    if (output)
       gzclose(output);
    free(palette);
    free(scanp);
}

#ifdef XPM
#include <xpm.h>

save_as_xpm_file (display, screen, win_info, pixmap, width, height, depth)
     Display *display;
     int screen;
     XWindowAttributes *win_info;
     Pixmap pixmap;
     unsigned int width, height, depth;
{
    XpmAttributes attributes;

    attributes.colormap = private_cmap;  /* win_info->colormap; */
    attributes.width = width;
    attributes.height = height;
    attributes.valuemask = XpmColormap | XpmSize;
    if (*root_name != '\0')
      check_filename();
    XpmWriteFileFromPixmap(display, 
      (*root_name)?output_name:NULL,pixmap,(Pixmap)0,&attributes);
}
#endif

#ifdef PNG
#include <png.h>

typedef struct _jmpbuf_wrapper {
  jmp_buf jmpbuf;
} jmpbuf_wrapper;

static jmpbuf_wrapper jmpbuf_struct;
static char *png_err_msg = "Fatal libpng error\n";

static void
png_error_handler (png_structp png_ptr, png_const_charp msg)
{
    jmpbuf_wrapper  *jmpbuf_ptr;

    fprintf(stderr, png_err_msg);
    fflush(stderr);

    jmpbuf_ptr = png_get_error_ptr(png_ptr);
    if (jmpbuf_ptr == NULL) {         /* we are completely hosed now */
        fprintf(stderr, "Extremely %s", png_err_msg);
        fflush(stderr);
        exit(1);
    }

    longjmp(jmpbuf_ptr->jmpbuf, 1);
}

save_as_png_file (display, screen, win_info, pixmap, width, height, depth)
     Display *display;
     int screen;
     XWindowAttributes *win_info;
     Pixmap pixmap;
     unsigned int width, height, depth;
{
    int           i, j, color_depth, ncolors;
    int           bit_depth=0, color_type=0, num_palette, rowbytes;
    png_colorp    palette = NULL;
    png_structp   png_ptr;
    png_infop     info_ptr;
    png_bytep     *row_pointers, png_data;
    char          *scanp, *ximp;
    Visual        *visual;
    XColor *colors;
    XImage *image;
    FILE *output;

    image = XGetImage (display, pixmap, 0,0, width,height, AllPlanes, ZPixmap);
    color_depth = image->bits_per_pixel;
    visual = DefaultVisual(display, screen);
    if (color_depth == 16 && visual->green_mask==992) color_depth = 15;
    if (ImageByteOrder(display) == MSBFirst) color_depth = -color_depth;

    if (!image) {
	fprintf (stderr, "xsnap:  unable to get image at %dx%d+%d+%d\n",
		 width, height, 0, 0);
	return;
    }

    /*
     * Find the output file
     */

    if (*root_name == '\0')
      output = stdout;
    else
      {
      check_filename();
      output = fopen (output_name, "wb");
      }
    if (!output) return;

    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
      &jmpbuf_struct, png_error_handler, NULL);
    if (!png_ptr) {
        fprintf(stderr, png_err_msg);
        fclose(output);
        return;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr) {
        fprintf(stderr, png_err_msg);
        png_destroy_read_struct(&png_ptr, NULL, NULL);
        fclose(output);
        return;
    }

    if (setjmp(jmpbuf_struct.jmpbuf)) {
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(output);
        return;
    }

    png_init_io(png_ptr, output);

    color_type = -1;  /* quiet compiler warning (all cases actually covered) */
    bit_depth = -1;

    if (image->depth == 1) {
        color_type = PNG_COLOR_TYPE_GRAY;
        bit_depth = 1;
    } else if (image->depth == 8) {
        char *ptr;
        color_type = PNG_COLOR_TYPE_PALETTE;
        bit_depth = 8;
        win_info->colormap = private_cmap;
        ncolors = Get_XColors(display, win_info, &colors);
        ptr = malloc(3*ncolors);
        palette = (png_colorp)ptr;
        if (!palette) {
	   fprintf(stderr, err_malloc, "palette");
           png_destroy_write_struct(&png_ptr, &info_ptr);
           fclose(output);
           return;
	}
        for (i=0;i<ncolors;i++) {
	   *ptr++ = colors[i].red>>8;
	   *ptr++ = colors[i].green>>8;
	   *ptr++ = colors[i].blue>>8;
	}
        png_set_PLTE(png_ptr, info_ptr, palette, ncolors);
    } else if (image->depth > 8) {
        color_type = PNG_COLOR_TYPE_RGB;
        bit_depth = 8;
    }

    png_set_IHDR(png_ptr, info_ptr, image->width, image->height, bit_depth,
      color_type, 0, PNG_COMPRESSION_TYPE_DEFAULT,
      PNG_FILTER_TYPE_DEFAULT);

    /* store the modification time, at least */
    {
        png_time  modtime;
        png_convert_from_time_t(&modtime, time(NULL));
        png_set_tIME(png_ptr, info_ptr, &modtime);
    }

    /* only one text comment:  Software */
    {
        png_text textdata;
        char software_text[40];
        sprintf(software_text, "Xsnap %s.%s", VERSION, PATCHLEVEL);
        textdata.compression = PNG_TEXT_COMPRESSION_NONE;
        textdata.key = "Software";
        textdata.text = software_text;
        png_set_text(png_ptr, info_ptr, &textdata, 1);
    }

    /* write all chunks up to (but not including) first IDAT */
    png_write_info(png_ptr, info_ptr);
    png_write_flush(png_ptr);

    /* any transformations must be set *after* png_write_info() is called */
    png_set_packing(png_ptr);   /* squish multiple pixels into each byte */
/*  png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);  */

    rowbytes = png_get_rowbytes(png_ptr, info_ptr);

    png_data = (png_bytep) malloc(image->height*rowbytes);
    row_pointers = (png_bytep *)malloc(image->height * sizeof(png_bytep));

    if (!png_data || !row_pointers) {
        fprintf(stderr, err_malloc, "png buffer");
        free(png_data);
        free(row_pointers);
        fprintf(stderr, png_err_msg);
        fflush(stderr);
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(output);
        return;
    }

    for (j = 0;  j < image->height;  ++j) {
        scanp = row_pointers[j] = (png_bytep)png_data + j*rowbytes;
        ximp = image->data + j * image->bytes_per_line;
        scan_ximage_line(scanp, ximp, image->width, color_depth);
    }

    png_write_image(png_ptr, row_pointers);
    free(row_pointers);
    free(png_data);
    free(palette);

    png_write_end(png_ptr, NULL);

    png_destroy_write_struct(&png_ptr, &info_ptr);
    if (output!=stdout)
        fclose(output);
}
#endif

#ifdef JPG

#include <setjmp.h>
#include <jpeglib.h>

struct error_mgr {
    struct jpeg_error_mgr pub;	/* "public" fields */
    jmp_buf setjmp_buffer;	/* for return to caller */
};

static struct error_mgr jerr;

typedef struct error_mgr *error_ptr;

static void
error_exit(j_common_ptr cinfo)
{
    char buf[JMSG_LENGTH_MAX];
    /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
    error_ptr err = (error_ptr) cinfo->err;

    cinfo->err->format_message(cinfo, buf);
    fprintf(stderr, "%\n", buf);

    /* Return control to the setjmp point */
    longjmp(err->setjmp_buffer, 1);
}

save_as_jpg_file (display, screen, win_info, pixmap, width, height, depth)
     Display *display;
     int screen;
     XWindowAttributes *win_info;
     Pixmap pixmap;
     unsigned int width, height, depth;
{
    int           i, j, color_depth, ncolors;
    int           bit_depth=0, color_type=0, num_palette, rowbytes;
    unsigned char * palette = NULL;
    unsigned char *scanp=NULL, *ximp;
    char          outstr[60];
    Visual        *visual;
    XColor *colors;
    XImage *image;
    struct jpeg_compress_struct cinfo;
    FILE *output;
    JSAMPROW scanline[1];

    image = XGetImage (display, pixmap, 0,0, width,height, AllPlanes, ZPixmap);
    color_depth = image->bits_per_pixel;
    visual = DefaultVisual(display, screen);
    if (color_depth == 16 && visual->green_mask==992) color_depth = 15;
    if (ImageByteOrder(display) == MSBFirst) color_depth = -color_depth;

    if (!image) {
        fprintf (stderr, "xsnap:  unable to get image at %dx%d+%d+%d\n",
		 width, height, 0, 0);
	return;
    }

    /*
     * Find the output file
     */

    if (*root_name == '\0')
      output = stdout;
    else
      {
      check_filename();
      output = fopen (output_name, "wb");
      if (!output) return;
      }

    if (image->depth == 8) {
        char *ptr;
        win_info->colormap = private_cmap;
        ncolors = Get_XColors(display, win_info, &colors);
        palette = ptr = (unsigned char *)malloc(3*ncolors);
        if (!palette) {
	  fprintf (stderr, err_malloc, "palette");
	  if (output)
              gzclose(output);
           return;
	}
        for (i=0;i<ncolors;i++) {
	   *ptr++ = colors[i].red>>8;
	   *ptr++ = colors[i].green>>8;
	   *ptr++ = colors[i].blue>>8;
	}
    }

    cinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = error_exit;

    if (setjmp(jerr.setjmp_buffer)) {
	/* If we get here, the JPEG code has signaled an error.
	 * We need to clean up the JPEG object, close the output file,
	 * and return.
	 */
	jpeg_destroy_compress(&cinfo);
	fclose(output);
	return 1;
    }
    jpeg_create_compress(&cinfo);
    jpeg_stdio_dest(&cinfo, output);

    cinfo.image_height = image->height;
    cinfo.image_width = image->width;
    cinfo.in_color_space = JCS_RGB;
    cinfo.input_components = 3;

    jpeg_set_defaults(&cinfo);
    jpeg_start_compress(&cinfo, TRUE);

    scanp = (unsigned char *)malloc(3*image->width);
    if (!scanp) {
        fprintf (stderr, err_malloc, "scanline");
        free(palette);
	if (output)
            gzclose(output);
        return;
    }

    scanline[0] = scanp;    
    for (j = 0;  j < image->height;  ++j) {
        ximp = image->data + j * image->bytes_per_line;
	if (image->depth==8) {
	   i = image->width-1;
           while(i>=0) {
              memcpy(&scanp[3*i], &palette[3*ximp[i]], 3);
              --i;
	   }
 	} else
           scan_ximage_line(scanp, ximp, image->width, color_depth);
	jpeg_write_scanlines(&cinfo, scanline, (JDIMENSION) 1);
    }
    jpeg_finish_compress(&cinfo);
    jpeg_destroy_compress(&cinfo);

    if (output!=stdout)
       fclose(output);

    free(palette);
    free(scanp);
    if (jerr.pub.num_warnings > 0) {	/* XXX */
	longjmp(jerr.setjmp_buffer, 1);
    }
}
#endif

void
save_image_file(display, screen, win_to_snap, pixmap, width, height, depth)
     Display *display;
     int screen;
     Window win_to_snap;
     Pixmap pixmap;
     unsigned int width, height, depth;
{
XWindowAttributes window_attributes;

        XGetWindowAttributes (display, win_to_snap, &window_attributes);
	if (output_fmt==PNM_F || output_fmt==PPM_F)
	  {
	    save_as_ppm_file (display, screen, &window_attributes, 
			      pixmap, width, height, depth);
	  }

	if (output_fmt==PNG_F)
	  {
#ifdef PNG
	    save_as_png_file (display, screen, &window_attributes,
                pixmap, width, height, depth);
#else
	    fprintf(stderr, "This version of xsnap has be compiled without PNG support,\n");
	    fprintf(stderr, "recompile with the -DPNG flag set.\n");
#endif
	  }

	if (output_fmt==JPG_F)
	  {
#ifdef JPG
	    save_as_jpg_file (display, screen, &window_attributes,
                pixmap, width, height, depth);
#else
	    fprintf(stderr, "This version of xsnap has be compiled without JPG support,\n");
	    fprintf(stderr, "recompile with the -DJPG flag.\n");
#endif
	  }

	if (output_fmt==XPM_F)
	  {
#ifdef XPM
	    save_as_xpm_file (display, screen, &window_attributes, 
                pixmap, width, height, depth);
#else
	    fprintf(stderr, "This version of xsnap has be compiled without XPM support,\n");
	    fprintf(stderr, "recompile with the -DXPM flag set.\n");
#endif
	  }

	if (output_fmt==XWD_F)
	{
	    save_as_xwd_file (display, screen, &window_attributes, 
                pixmap, width, height, depth);
	}
}

exitXsnap(display, snapshot, dialog_window, snap_pixmap, copy_gc, restart)
     Display *display;
     Window snapshot;
     Window dialog_window;
     Pixmap snap_pixmap;
     GC copy_gc;
     int restart;
{
  if (snapshot != 0)
    XDestroyWindow(display, snapshot);
  if (dialog_window != 0)
    XDestroyWindow(display, dialog_window);
  XFreePixmap(display, snap_pixmap);
  XFreeGC(display, copy_gc);
  XCloseDisplay(display);
  if (restart) system("xsnap &");
  exit(0);
}

void
draw_dialog(Display *dpy, Window win, GC gc, int u, int v)
{
  int i, x, y;  
  static int oldzip = -1;
  static int i0 = -1;
  static char old_name[256] = "/";
  char msg[512];
  char *ptr1, *ptr2;
  for (i=1; i<=3; i++) {
     ptr1 = bindings_msg[i] + 1;
     XDrawString(dpy, win, gc, 10, 20*i, ptr1, strlen(ptr1));  
  }
  for (i=0; i<9; i++) {
     if (i==i0) {
       XClearArea(dpy, win, 11, 83+30*i, 82, 2, FALSE);
       XClearArea(dpy, win, 11, 102+30*i, 82, 2, FALSE);
       XClearArea(dpy, win, 11, 83+30*i, 2, 20, FALSE);
       XClearArea(dpy, win, 92, 83+30*i, 2, 20, FALSE);
     }
     XSetForeground(dpy, gc, color_pixel[(i==clicked)?BT1_COLOR:BT3_COLOR]);
     XDrawLine(dpy, win, gc, 9, 81+30*i, 9, 105+30*i);
     XDrawLine(dpy, win, gc, 9, 81+30*i, 95, 81+30*i);
     XSetForeground(dpy, gc, color_pixel[(i==clicked)?BT2_COLOR:BT4_COLOR]);
     XDrawLine(dpy, win, gc, 10, 82+30*i, 10, 104+30*i);
     XDrawLine(dpy, win, gc, 10, 82+30*i, 94, 82+30*i);
     XSetForeground(dpy, gc, color_pixel[(i==clicked)?BT4_COLOR:BT2_COLOR]);
     XDrawLine(dpy, win, gc, 10, 104+30*i, 94, 104+30*i);
     XDrawLine(dpy, win, gc, 94, 82+30*i, 94, 104+30*i);
     XSetForeground(dpy, gc, color_pixel[(i==clicked)?BT3_COLOR:BT1_COLOR]);
     XDrawLine(dpy, win, gc,  9, 105+30*i, 95, 105+30*i);
     XDrawLine(dpy, win, gc, 95, 81+30*i, 95, 105+30*i);
     XSetForeground(dpy, gc, color_pixel[FG_COLOR]);
     if (u>10 && u<94 && v>82+30*i && v<104+30*i) {
       i0 = i;
       for(y=84+30*i; y<=102+30*i; y+=18)
	 for(x=12; x<=92; x++) if ((x+y)&1) XDrawPoint(dpy, win, gc, x, y);
       for(y=84+30*i; y<=102+30*i; y++)
	 for(x=12; x<=92; x+=80) if ((x+y)&1) XDrawPoint(dpy, win, gc, x, y);
     }
     ptr1 = bindings_msg[i+6]+1;
     ptr2 = index(ptr1, '\t');
     XDrawString(dpy, win, gc, 40, 98+30*i, ptr1, ptr2-ptr1); 
     ++ptr2;
     XDrawString(dpy, win, gc, 140, 98+30*i, ptr2, strlen(ptr2));  
  }
  ptr1 = (zip)?"[active]":"[inactive]"; 
  if (zip!=oldzip) XClearArea(dpy, win, 340, 110, 200, 40, FALSE);
  oldzip = zip;
  XDrawImageString(dpy, win, gc, 340, 128, ptr1, strlen(ptr1));
  if (strcmp(old_name,output_name) || (u==0 && v==0)) {
     XClearArea(dpy, win, 0, 366, 450, 30, FALSE);
     if (*output_name)
       sprintf(msg, "Image saved in  %s", output_name);
     else
       strcpy(msg, "Image not saved");
     XDrawString(dpy, win, gc, 10, 378, msg, strlen(msg));
     strcpy(old_name,output_name);
  }
  XFlush(dpy);
}

int
main(argc, argv)
int argc;
char **argv;
{
	Display *display;
	int screen;
	XWMHints wmhints;
	XEvent ev;
        XFontStruct * font;
	GC copy_gc;
        Atom wm_delete_window, wm_protocols;
	unsigned long copy_gc_valuemask;
	XGCValues copy_gc_values;
	long wmhints_mask;
	XSizeHints wm_size_hints;
	int x_return, y_return;           /* for XParseGeometry */
	int width_return, height_return;  /* ditto              */
	int geom_mask;                    /* bitmask for XParseGeometry fields */
	XColor color;
        KeySym keysym = 0;
	Colormap cmap;
	Window window_to_snap;            /* always root window, now */
	Window childW;                    /* dummy window variable */
	Window snapshot = 0;
   	Window dialog_window = 0;
	Cursor snap_cursor;
	Pixmap  snap_pixmap;
	int i, i0=-10, reg_x, reg_y;
	int icon_x, icon_y;
	int snap_x, snap_y;
	unsigned int snap_width, snap_height;
	unsigned int reg_width, reg_height, reg_depth;
	int done;
	char buffer [4];                     /* key press buffer for XLookupString  */
	int string_length;                       /* length of returned string */
	XWindowAttributes window_attributes;

	process_args(argc, argv);
	XSetErrorHandler(my_X_error_handler);
	if ((display = XOpenDisplay(display_string)) == NULL)
	  {
	    fprintf(stderr, "Cannot open display:    %s\n", 
		    XDisplayName(display_string));
	    exit(1);
	  }
	XSetErrorHandler(my_X_error_handler);
	screen = DefaultScreen(display);
	window_to_snap = XRootWindow(display, screen);
	font = XLoadQueryFont(display, "7x13");

        /* WM protocols */
        wm_protocols = XInternAtom(display, "WM_PROTOCOLS", False);
        wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", False);

	/* process colors */
 
        if (private_cmap && DefaultDepth(display, screen)<15)
            cmap = private_cmap;
        else
	    cmap = DefaultColormap(display, screen);

        for (i=0; i<NUM_COLORS; i++) {
            if ((color_string[i]) && 
                (XParseColor(display, cmap, color_string[i], &color)) &&
	        (XAllocColor(display, cmap, &color)))
	        color_pixel[i] = color.pixel;
	    else
	        color_pixel[i] = (i==BG_COLOR)?
	            WhitePixel(display, screen):BlackPixel(display, screen);
	}

	/* make copy GC */

	copy_gc_valuemask = GCSubwindowMode | GCFont |
                            GCForeground | GCBackground;
	copy_gc_values.subwindow_mode = IncludeInferiors;
	copy_gc_values.background = color_pixel[BG_COLOR];
	copy_gc_values.foreground = color_pixel[FG_COLOR];
	copy_gc_values.font = font->fid;
	copy_gc = XCreateGC(display, window_to_snap, copy_gc_valuemask, &copy_gc_values);

	if (grab_server)
	{
		XGrabServer(display);
		XSync(display, 0);
	}

	if (region_geom_string)
	{
		geom_mask = XParseGeometry(region_geom_string, &reg_x, &reg_y,
		    &reg_width, &reg_height);
		if ((geom_mask & XValue) && (geom_mask & YValue) &&
		    (geom_mask & WidthValue) && (geom_mask & HeightValue))
		/* must specify complete geometry for region */
		{
			if (geom_mask & XNegative)
				reg_x += DisplayWidth(display, screen) - reg_width;
			if (geom_mask & YNegative)
				reg_y += DisplayHeight(display, screen) - reg_height;
			snap_pixmap = get_pixmap_region(display, screen, 
			    window_to_snap, copy_gc,
			    &reg_x, &reg_y, &reg_width, 
			    &reg_height, &reg_depth, FALSE);
		}
		else {
		    if (grab_server) {
			XUngrabServer(display);
			XSync(display, 0);
		    }
		    usage();
		}
	}
	else 
		snap_pixmap = get_pixmap_region(display, screen,
						window_to_snap, copy_gc,
						&reg_x, &reg_y,
						&reg_width, &reg_height,
						&reg_depth, TRUE);

	/* ungrab the server */

	if (grab_server) XUngrabServer(display);

	if (display_in_window)
	  {
	    /*
	     * get location of our window(from the window_geom_string),
	     * and set up the size hints structure.
	     */     
	    wm_size_hints.flags = 0;
	    width_return = reg_width;
	    height_return = reg_height;
	    wm_size_hints.x = 0;
	    wm_size_hints.y = 0;
	    wm_size_hints.width = reg_width;
	    wm_size_hints.height = reg_height;

	    if (window_geom_string)
	      {
		geom_mask = XParseGeometry(window_geom_string, &x_return, 
					   &y_return, &width_return, 
					   &height_return);
		if (geom_mask & XValue)
		  {
		    if (geom_mask & XNegative)
		      wm_size_hints.x = DisplayWidth(display, 
						     screen)
			- width_return + x_return - 
			  (border_width * 2);
		    else
		      wm_size_hints.x = x_return;
		  }
		if (geom_mask & YValue)
		  {
		    if (geom_mask & YNegative)
		      wm_size_hints.y =  
			DisplayHeight(display, screen)
			  - height_return + y_return  - 
			    (border_width * 2);

		    else
		      wm_size_hints.y = y_return;
		  }

	      }

	    if ((geom_mask & XValue) || (geom_mask & YValue))
	      wm_size_hints.flags |= USPosition;

	    if ((geom_mask & WidthValue) || (geom_mask & HeightValue))
	      wm_size_hints.flags |= USSize;

            wm_size_hints.flags |=  PMaxSize | PMinSize;
	    wm_size_hints.width = width_return;
	    wm_size_hints.max_width = width_return;
	    wm_size_hints.min_width = width_return;
	    wm_size_hints.height = height_return;
	    wm_size_hints.max_height = height_return;
	    wm_size_hints.min_height = height_return;
	    
	    snapshot = XCreateSimpleWindow(display, 
					   XRootWindow(display, screen), 
					   wm_size_hints.x, wm_size_hints.y,
					   wm_size_hints.width, 
					   wm_size_hints.height, border_width,
					   color_pixel[BD_COLOR],
					   color_pixel[BG_COLOR]);

	    XSetWindowColormap(display, snapshot, private_cmap);
            XSetWMProtocols(display, snapshot, &wm_delete_window, True);
            XSetNormalHints(display, snapshot, &wm_size_hints);

    	    dialog_window = XCreateSimpleWindow(display, 
		XRootWindow(display, screen), 
		0, 0, 450, 392, border_width,
	        color_pixel[BD_COLOR],
		color_pixel[BG_COLOR]);

	    XSetWindowColormap(display, dialog_window, private_cmap);
  	    XStoreName(display, dialog_window, "xsnap / menu");
            XSetWMProtocols(display, dialog_window, 
                            &wm_delete_window, True);

	    if (window_geom_string)
	      XSetNormalHints(display, snapshot, &wm_size_hints);
	    wmhints_mask = 0;

	    /* set the input hint to true explicitly for OpenWindows */
	    wmhints_mask |= InputHint;
	    wmhints.input = True;

	    if (start_iconic)
	      {
		wmhints_mask |= StateHint;
		wmhints.initial_state = IconicState;
	      }
	    x_return = y_return = 0;

	    /* Icon geometry is sorta broken, since we don't supply an icon, and 
	     * there isn't any way I know of to get the geometry  of the icon that
	     * the window manager will provide.   So, we'll pretend that our icon 
	     * is 32x32 (ugh).
	     */

	    if (icon_geom_string)
	      {
		geom_mask = XParseGeometry(icon_geom_string, &x_return, 
					   &y_return, &width_return, 
					   &height_return);
		if (geom_mask & XValue)
		  {
		    if (geom_mask & XNegative)
		      wmhints.icon_x = DisplayWidth(display, 
						    screen)
			- 32 + x_return - (border_width * 2);
		    else
		      wmhints.icon_x = x_return;
		  }
		if (geom_mask & YValue)
		  {
		    if (geom_mask & YNegative)
		      wmhints.icon_y =  DisplayHeight(display, 
						      screen)
			- 32 + y_return  - (border_width * 2);

		    else
		      wmhints.icon_y = y_return;
		  }
		wmhints_mask |= IconPositionHint;
	      }


	    if (wmhints_mask)
	      {
		wmhints.flags = wmhints_mask;
		XSetWMHints(display, snapshot, &wmhints);
	      }

	    /* sollicit expose events */

	    XSelectInput(display, snapshot, 
                ExposureMask | KeyPressMask |
                ButtonPressMask | ButtonReleaseMask);

	    XSelectInput(display, dialog_window, 
                ExposureMask | KeyPressMask | PointerMotionMask |
                ButtonPressMask | ButtonReleaseMask);

	    /* give it a different cursor */

	    snap_cursor = XCreateFontCursor(display, XC_arrow);

	    XDefineCursor(display, snapshot, snap_cursor);

	    /* set the background of the window to be the pixmap */

	    XSetWindowBackgroundPixmap (display, snapshot, snap_pixmap);
	}


       /* map it */

	if (display_in_window) {
            XMapRaised(display, snapshot);   
	    XFlush (display);
	}

	if (!*win_name0) 
            strcpy(win_name0, "xsnap");

	sprintf(win_name, "%s %d x %d", 
                          win_name0, width_return, height_return);
	XStoreName(display, snapshot, win_name);

    save_file:
        if (save_in_file) {
	    save_image_file(display, screen, window_to_snap,
		 	   snap_pixmap, reg_width, reg_height, reg_depth);
	    if (!root_name) 
                strcpy(output_name, "(stdout)");
	    sprintf(win_name, "%s %d x %d -> %s", 
                            win_name0, width_return, height_return,
                            output_name);
	    XStoreName(display, snapshot, win_name);
	    draw_dialog(display, dialog_window, copy_gc,
                                 50, 96+30*i0);
	}

	if (!display_in_window)
	    exitXsnap (display, NULL, NULL, snap_pixmap, copy_gc, 0);
   
	/* process events */
	done = FALSE;

	while (display_in_window && !done)
	{
		XNextEvent(display, &ev);
		switch (ev.type)
		{
		case ClientMessage:
		        if (ev.xclient.message_type == wm_protocols && 
                            ev.xclient.format == 32 &&
                            ev.xclient.data.l[0] == wm_delete_window) {
			    if (ev.xexpose.window == dialog_window)
			        XUnmapWindow(display, dialog_window);
			    if (ev.xexpose.window == snapshot)
			        exitXsnap(display, snapshot, 
                                     dialog_window, snap_pixmap, copy_gc, 0);
			}
                        break;
		case MotionNotify:
		        if (ev.xexpose.window == dialog_window) {
			    draw_dialog(display, dialog_window, copy_gc,
                                        ev.xbutton.x, ev.xbutton.y);
			}
			break;
		case Expose:
		case MappingNotify:
		        if (ev.xexpose.window == dialog_window) {
			    draw_dialog(display, dialog_window, copy_gc, 0, 0);
			    break;
		        }
			XRefreshKeyboardMapping((XMappingEvent *)&ev);
			break;
		case ButtonPress:
		case ButtonRelease:
		        if (ev.xexpose.window == dialog_window) {
			    i0 = -1;
			    for (i=0; i<9; i++)
                               if (ev.xbutton.x>10 && ev.xbutton.x<94 && 
                                   ev.xbutton.y>82+30*i && 
                                   ev.xbutton.y<104+30*i) {
				   i0 = i;
				   break;
			       }
			    if (ev.type==ButtonRelease &&
                                (i0==-1 || i0!=clicked)) {
			        clicked = -1;
			        draw_dialog(display, dialog_window, 
                                            copy_gc, 
					    ev.xbutton.x, ev.xbutton.y);
                                break;
			    }
			    clicked = (ev.type==ButtonPress)? i0:-1;
			    draw_dialog(display, dialog_window, 
                                        copy_gc, 
					ev.xbutton.x, ev.xbutton.y);
                            if (ev.type==ButtonPress) break;
			    clicked = -1;
			    if (i0== 0) {
			        exitXsnap(display, snapshot, 
                                    dialog_window, snap_pixmap, copy_gc, 0);
			        done = TRUE;
			    }
			    if (i0== 1) {
			        zip = 1-zip;
			        draw_dialog(display, dialog_window,
                                            copy_gc, 50, 126);
				break;
			    }
			    i0 -= 2;
			    if (i0>=JPG_F && i0<NUM_F) {
			        output_fmt = i0;
                                save_in_file = TRUE;
				i0 += 2;
			        goto save_file;
			    }
			    if (i0 == NUM_F) {
			        exitXsnap(display, snapshot, 
                                    dialog_window, snap_pixmap, copy_gc, 1);
			        done = TRUE;
			    }
			    break;
		        }
		   raise_dialog:
			clicked = -1;
                        wm_size_hints.flags = PPosition;
                        XTranslateCoordinates(display, snapshot,
			    XRootWindow(display, screen),
                            0, 0, &snap_x, &snap_y, 
                            &childW);
	                wm_size_hints.x = snap_x;
 	                wm_size_hints.y = snap_y;
			if (ev.xbutton.button==3) {
	                    XMapRaised(display, dialog_window);
                            XMoveWindow(display, dialog_window, 
                                        snap_x+ev.xbutton.x-10, 
                                        snap_y+ev.xbutton.y-10);
			}
			break;
		case KeyPress:
			string_length = XLookupString((XKeyEvent *)&ev, 
                            buffer, sizeof(buffer), &keysym, NULL);
                        buffer[0] = tolower(keysym);
			if (buffer[0] == 'q' || buffer[0] == '\033')
			    {
			    /*  clean up  */
			    exitXsnap(display, snapshot, 
                               dialog_window, snap_pixmap, copy_gc, 0);
			    done = TRUE;
			    }
                        else
			if (buffer[0] == '!')
			    {
			    exitXsnap(display, snapshot, 
                               dialog_window, snap_pixmap, copy_gc, 1);
			    done = TRUE;
			    }
                        else
			if (buffer[0] == 'j')
			    {
			    output_fmt = JPG_F;
                            save_in_file = TRUE;
			    i0 = output_fmt+2;
			    goto save_file;
			    }
                        else
			if (buffer[0] == 'n')
			    {
			    output_fmt = PNM_F;
                            save_in_file = TRUE;
			    i0 = output_fmt+2;
			    goto save_file;
			    }
                        else
			if (buffer[0] == 'p')
			    {
			    output_fmt = PPM_F;
                            save_in_file = TRUE;
			    i0 = output_fmt+2;
			    goto save_file;
			    }
                        else
			if (buffer[0] == 'g')
			    {
			    output_fmt = PNG_F;
                            save_in_file = TRUE;
			    i0 = output_fmt+2;
			    goto save_file;
			    }
                        else
			if (buffer[0] == 'x')
			    {
			    output_fmt = XPM_F;
                            save_in_file = TRUE;
			    i0 = output_fmt+2;
			    goto save_file;
			    }
                        else
			if (buffer[0] == 'w')
			    {
			    output_fmt = XWD_F;
                            save_in_file = TRUE;
			    i0 = output_fmt+2;
			    goto save_file;
			    }
                        else
			if (buffer[0] == 'z')
			    {
			    zip = 1-zip;
			    draw_dialog(display, dialog_window, copy_gc, 
                                        50, 130);
			    }
                        else {
			  if (!isalpha(keysym) && !isspace(keysym)) break;
			  next:
			    ev.xbutton.button = 3;
			    ev.xbutton.x = 20;
			    ev.xbutton.y = 20;
			    goto raise_dialog;
			}
			break;
		      }
	}
	return 0;
}

